unit AcadEvents;

interface

uses
  Windows, Classes, Variants, ComObj, ActiveX, OleServer, AutoCAD_TLB;

type
  TAcadWindowMovedEvent = procedure(Sender: TObject; AHandle: Integer; AMoved: WordBool) of object;
  TAcadBeginCloseEvent = procedure(Sender: TObject; var ACancel: WordBool) of object;
  TAcadCommandEvent = procedure(Sender: TObject; const ACommand: WideString) of object;
  TAcadPointEvent = procedure(Sender: TObject; APoint: OleVariant) of object;
  TAcadObjectEvent = procedure(Sender: TObject; AObject: OleVariant) of object;
  TAcadEditEvent = procedure(Sender: TObject; AObject: OleVariant; AParam: OleVariant) of object;
  TAcadEditCommandEvent = procedure(Sender: TObject; AObject: OleVariant; const ACommand: WideString) of object;
  TAcadErasedEvent = procedure(Sender: TObject; AObjectID: Integer) of object;
  TAcadWindowChangedEvent = procedure(Sender: TObject; AState: TOleEnum) of object;
  TAcadUnknownEvent = procedure(Sender: TObject; ADispID: TDispID) of object;

  TAcadDocumentEvents = class(TObject, IUnknown, IDispatch)
  private
    FOwner: TObject;
    FOnBeginSave: TAcadCommandEvent;
    FOnEndSave: TAcadCommandEvent;
    FOnBeginCommand: TAcadCommandEvent;
    FOnEndCommand: TAcadCommandEvent;
    FOnBeginLisp: TAcadCommandEvent;
    FOnEndLisp: TNotifyEvent;
    FOnLispCancelled: TNotifyEvent;
    FOnSelectionChanged: TNotifyEvent;
    FOnActivate: TNotifyEvent;
    FOnDeactivate: TNotifyEvent;
    FOnBeginRightClick: TAcadPointEvent;
    FOnBeginShortcutMenuDefault: TAcadObjectEvent;
    FOnBeginShortcutMenuEdit: TAcadEditEvent;
    FOnBeginShortcutMenuCommand: TAcadEditCommandEvent;
    FOnBeginShortcutMenuGrip: TAcadObjectEvent;
    FOnBeginShortcutMenuOsnap: TAcadObjectEvent;
    FOnEndShortcutMenu: TAcadObjectEvent;
    FOnBeginDoubleClick: TAcadPointEvent;
    FOnObjectAdded: TAcadObjectEvent;
    FOnObjectErased: TAcadErasedEvent;
    FOnObjectModified: TAcadObjectEvent;
    FOnBeginPlot: TAcadCommandEvent;
    FOnEndPlot: TAcadCommandEvent;
    FOnWindowMovedOrResized: TAcadWindowMovedEvent;
    FOnLayoutSwitched: TAcadCommandEvent;
    FOnWindowChanged: TAcadWindowChangedEvent;
    FOnBeginClose: TNotifyEvent;
    FOnBeginDocClose: TAcadBeginCloseEvent;
    FOnUnknownEvent: TAcadUnknownEvent;
  protected
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    function GetTypeInfoCount(out Count: Integer): HResult; virtual; stdcall;
    function GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; virtual; stdcall;
    function GetIDsOfNames(const IID: TGUID; Names: Pointer;
      NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; virtual; stdcall;
    function Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer;
      Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; virtual; stdcall;
    procedure InvokeEvent(DispID: TDispID; var Params: TVariantArray);
  public
    constructor Create(Owner: TObject);
    property OnBeginSave: TAcadCommandEvent read FOnBeginSave write FOnBeginSave;
    property OnEndSave: TAcadCommandEvent read FOnEndSave write FOnEndSave ;
    property OnBeginCommand: TAcadCommandEvent read FOnBeginCommand write FOnBeginCommand;
    property OnEndCommand: TAcadCommandEvent read FOnEndCommand write FOnEndCommand;
    property OnBeginLisp: TAcadCommandEvent read FOnBeginLisp write FOnBeginLisp ;
    property OnEndLisp: TNotifyEvent read FOnEndLisp write FOnEndLisp;
    property OnLispCancelled: TNotifyEvent read FOnLispCancelled write FOnLispCancelled;
    property OnSelectionChanged: TNotifyEvent read FOnSelectionChanged write FOnSelectionChanged;
    property OnActivate: TNotifyEvent read FOnActivate write FOnActivate;
    property OnDeactivate: TNotifyEvent read FOnDeactivate write FOnDeactivate;
    property OnBeginRightClick: TAcadPointEvent read FOnBeginRightClick write FOnBeginRightClick;
    property OnBeginShortcutMenuDefault: TAcadObjectEvent read FOnBeginShortcutMenuDefault write FOnBeginShortcutMenuDefault;
    property OnBeginShortcutMenuEdit: TAcadEditEvent read FOnBeginShortcutMenuEdit write FOnBeginShortcutMenuEdit;
    property OnBeginShortcutMenuCommand: TAcadEditCommandEvent read FOnBeginShortcutMenuCommand write FOnBeginShortcutMenuCommand;
    property OnBeginShortcutMenuGrip: TAcadObjectEvent read FOnBeginShortcutMenuGrip write FOnBeginShortcutMenuGrip;
    property OnBeginShortcutMenuOsnap: TAcadObjectEvent read FOnBeginShortcutMenuOsnap write FOnBeginShortcutMenuOsnap;
    property OnEndShortcutMenu: TAcadObjectEvent read FOnEndShortcutMenu write FOnEndShortcutMenu;
    property OnBeginDoubleClick: TAcadPointEvent read FOnBeginDoubleClick write FOnBeginDoubleClick;
    property OnObjectAdded: TAcadObjectEvent read FOnObjectAdded write FOnObjectAdded;
    property OnObjectErased: TAcadErasedEvent read FOnObjectErased write FOnObjectErased;
    property OnObjectModified: TAcadObjectEvent read FOnObjectModified write FOnObjectModified;
    property OnBeginPlot: TAcadCommandEvent read FOnBeginPlot write FOnBeginPlot;
    property OnEndPlot: TAcadCommandEvent read FOnEndPlot write FOnEndPlot;
    property OnWindowMovedOrResized: TAcadWindowMovedEvent read FOnWindowMovedOrResized write FOnWindowMovedOrResized;
    property OnLayoutSwitched: TAcadCommandEvent read FOnLayoutSwitched write FOnLayoutSwitched;
    property OnWindowChanged: TAcadWindowChangedEvent read FOnWindowChanged write FOnWindowChanged;
    property OnBeginClose: TNotifyEvent read FOnBeginClose write FOnBeginClose;
    property OnBeginDocClose: TAcadBeginCloseEvent read FOnBeginDocClose write FOnBeginDocClose;
    property OnUnknownEvent: TAcadUnknownEvent read FOnUnknownEvent write FOnUnknownEvent;
  end;

  function AcadConnect: Boolean;
  procedure AcadRelease;

var
  Acad: OleVariant;
  AcadDocEvents: TAcadDocumentEvents;
   
implementation

var
  CPC: IConnectionPointContainer;
  CP: IConnectionPoint;
  Cookie: Longint;
  IID_AcadDocumentEvents: TIID;

function AcadConnect: Boolean;
var
  Enum: IEnumConnectionPoints;
  N: Longint;
begin
  Result:= False;
  //   AutoCAD. ,  AutoCAD  !
  Acad:= GetActiveOleObject('AutoCAD.Application');
  //   
  if not VarIsClear(Acad) then
  begin
    //  IConnectionPointContainer
    if Succeeded(IUnknown(Acad.ActiveDocument).QueryInterface(IConnectionPointContainer, CPC)) then
    begin
      //  EnumConnectionPoints
      if Succeeded(CPC.EnumConnectionPoints(Enum)) then
      begin
        Enum.Next(1, CP, @N); //  IConnectionPoint
        if N = 1 then         //   !
        begin
          //      
          if Succeeded(CP.GetConnectionInterface(IID_AcadDocumentEvents)) then
            //    
            Result:= Succeeded(CP.Advise(AcadDocEvents, Cookie));
        end;
      end;
    end;
  end;
end;

procedure AcadRelease;
begin
  if not VarIsClear(Acad) then
  begin
    //    
    if Cookie <> 0 then
      CP.Unadvise(Cookie);
    Acad:= UnAssigned;
  end;
end;

{ TAcadDocumentEvents }

function TAcadDocumentEvents.GetTypeInfo(Index, LocaleID: Integer;
  out TypeInfo): HResult;
begin
  Pointer(TypeInfo):= nil;
  Result:= E_NOTIMPL;
end;

//         
//  IUnknown, IDispatch  IID_AcadDocumentEvents
function TAcadDocumentEvents.QueryInterface(const IID: TGUID;
  out Obj): HResult;
begin
  if GetInterface(IID, Obj) then
    Result:= S_OK
  else if IsEqualIID(IID, IID_AcadDocumentEvents) then
    Result:= QueryInterface(IDispatch, Obj)
  else
    Result:= E_NOINTERFACE;
end;

function TAcadDocumentEvents.GetIDsOfNames(const IID: TGUID; Names: Pointer;
  NameCount, LocaleID: Integer; DispIDs: Pointer): HResult;
begin
  Result:= E_NOTIMPL;
end;

function TAcadDocumentEvents._AddRef: Integer;
begin
  Result:= 1;
end;

function TAcadDocumentEvents.GetTypeInfoCount(out Count: Integer): HResult;
begin
  Count:= 0;
  Result:= S_OK;
end;

function TAcadDocumentEvents.Invoke(DispID: Integer; const IID: TGUID;
  LocaleID: Integer; Flags: Word; var Params; VarResult, ExcepInfo,
  ArgErr: Pointer): HResult;
var
  i: integer;
  vVarArray: TVariantArray;
  vPVarArgIn: PVariantArg;
  vPDispParams: PDispParams;
  vFistArrItem, vLastArrItem: integer;
begin
  vPDispParams:= PDispParams(@Params);     // DispParams
  SetLength(vVarArray, vPDispParams.cArgs); // set our array to appropriate length
  // array boundaries
  vFistArrItem:= Low(vVarArray); vLastArrItem:= High(vVarArray);
  if vPDispParams.cNamedArgs > 0 then
    // Copy over data from Params in NamedArgs order
    for i:= vFistArrItem to vLastArrItem do
    begin
      vPVarArgIn:= @vPDispParams.rgvarg[i];
      vVarArray[vPDispParams.rgdispidNamedArgs[i]]:= POleVariant(vPVarArgIn)^;
    end
  else
    // Copy over data from Params in reverse order
    for i:= vFistArrItem to vLastArrItem do
    begin
      vPVarArgIn:= @vPDispParams.rgvarg[i];
      vVarArray[vLastArrItem - i]:= POleVariant(vPVarArgIn)^;
    end;
  InvokeEvent(DispID, vVarArray);
  if vPDispParams.cNamedArgs > 0 then
    // Copy data back to DispParams if Item passed by reference (NamedArgs order)
    for i:= vFistArrItem to vLastArrItem do
    begin
      vPVarArgIn:= @vPDispParams.rgvarg[i];
      if (vPVarArgIn.vt and varByRef) <> varByRef then
        Continue;
      POleVariant(vPVarArgIn)^:= vVarArray[vPDispParams.rgdispidNamedArgs[i]];
    end
  else
    // Copy data back to DispParams if Item passed by reference (reverse order)
    for i:= vFistArrItem to vLastArrItem do
    begin
      vPVarArgIn:= @vPDispParams.rgvarg[i];
      if (vPVarArgIn.vt and varByRef) <> varByRef then
        Continue;
      POleVariant(vPVarArgIn)^:= vVarArray[vLastArrItem - i];
    end;
  // Clean array
  SetLength(vVarArray, 0);
  // Pascal Events return 'void' - so assume success!
  Result:= S_OK;
end;

function TAcadDocumentEvents._Release: Integer;
begin
  Result:= 1;
end;

constructor TAcadDocumentEvents.Create(Owner: TObject);
begin
  inherited Create;
  FOwner:= Owner;
end;

procedure TAcadDocumentEvents.InvokeEvent(DispID: TDispID;
  var Params: TVariantArray);
begin
  case DispID of
   -1: Exit;  // DISPID_UNKNOWN
    1: if Assigned(FOnBeginSave) then
         FOnBeginSave(Self, Params[0] {const WideString});
    2: if Assigned(FOnEndSave) then
         FOnEndSave(Self, Params[0] {const WideString});
    6: if Assigned(FOnBeginCommand) then
         FOnBeginCommand(Self, Params[0] {const WideString});
    7: if Assigned(FOnEndCommand) then
         FOnEndCommand(Self, Params[0] {const WideString});
    8: if Assigned(FOnBeginLisp) then
         FOnBeginLisp(Self, Params[0] {const WideString});
    9: if Assigned(FOnEndLisp) then
         FOnEndLisp(Self);
    10: if Assigned(FOnLispCancelled) then
         FOnLispCancelled(Self);
    11: if Assigned(FOnSelectionChanged) then
         FOnSelectionChanged(Self);
    12: if Assigned(FOnActivate) then
         FOnActivate(Self);
    13: if Assigned(FOnDeactivate) then
         FOnDeactivate(Self);
    14: if Assigned(FOnBeginRightClick) then
         FOnBeginRightClick(Self, Params[0] {OleVariant});
    15: if Assigned(FOnBeginShortcutMenuDefault) then
         FOnBeginShortcutMenuDefault(Self, IDispatch((TVarData(Params[0]).VPointer)^) {out IAcadPopupMenu});
    16: if Assigned(FOnBeginShortcutMenuEdit) then
         FOnBeginShortcutMenuEdit(Self,
                                  IDispatch((TVarData(Params[0]).VPointer)^) {out IAcadPopupMenu},
                                  IDispatch((TVarData(Params[1]).VPointer)^) {out IAcadSelectionSet});
    17: if Assigned(FOnBeginShortcutMenuCommand) then
         FOnBeginShortcutMenuCommand(Self,
                                     IDispatch((TVarData(Params[0]).VPointer)^) {out IAcadPopupMenu},
                                     Params[1] {const WideString});
    18: if Assigned(FOnBeginShortcutMenuGrip) then
         FOnBeginShortcutMenuGrip(Self, IDispatch((TVarData(Params[0]).VPointer)^) {out IAcadPopupMenu});
    19: if Assigned(FOnBeginShortcutMenuOsnap) then
         FOnBeginShortcutMenuOsnap(Self, IDispatch((TVarData(Params[0]).VPointer)^) {out IAcadPopupMenu});
    30: if Assigned(FOnEndShortcutMenu) then
         FOnEndShortcutMenu(Self, IDispatch((TVarData(Params[0]).VPointer)^) {out IAcadPopupMenu});
    20: if Assigned(FOnBeginDoubleClick) then
         FOnBeginDoubleClick(Self, Params[0] {OleVariant});
    21: if Assigned(FOnObjectAdded) then
         FOnObjectAdded(Self, Params[0] {const IDispatch});
    22: if Assigned(FOnObjectErased) then
         FOnObjectErased(Self, Params[0] {Integer});
    23: if Assigned(FOnObjectModified) then
         FOnObjectModified(Self, Params[0] {const IDispatch});
    24: if Assigned(FOnBeginPlot) then
         FOnBeginPlot(Self, Params[0] {const WideString});
    25: if Assigned(FOnEndPlot) then
         FOnEndPlot(Self, Params[0] {const WideString});
    29: if Assigned(FOnWindowMovedOrResized) then
         FOnWindowMovedOrResized(Self,
                                 Params[0] {Integer},
                                 Params[1] {WordBool});
    31: if Assigned(FOnLayoutSwitched) then
         FOnLayoutSwitched(Self, Params[0] {const WideString});
    32: if Assigned(FOnWindowChanged) then
         FOnWindowChanged(Self, Params[0] {AcWindowState});
    33: if Assigned(FOnBeginClose) then
         FOnBeginClose(Self);
    34: if Assigned(FOnBeginDocClose) then
         FOnBeginDocClose(Self, WordBool((TVarData(Params[0]).VPointer)^) {var WordBool});
  else
    if Assigned(FOnUnknownEvent) then
      FOnUnknownEvent(Self, DispID);
  end; {case DispID}
end;

end.
